import os
import shutil
import json
import tempfile
from isqtools import IsqCircuit
from isqtools.backend import NumpyBackend
import numpy as np
import os, json, sys, logging
from multiprocessing import Pool


try:
    sys.path.append(os.path.join(os.environ["PROJECTS"], "20230113_QumPy"))
    import Circuit
except:
    pass


# all simulator should input "so" filepath


# used by simulator
backend = NumpyBackend()


class findSimulator:
    @staticmethod
    def set_simulator_fileName():
        return "SIMULATOR"

    @staticmethod
    def get_dirName_of_Simulator(isqcDIR):
        storeDIR = os.path.join(isqcDIR, "nix", "store")
        messList = os.listdir(storeDIR)
        for l in messList:
            if "simulator" in l:
                return l.strip()
        logging.error("cannot find simulator dir")

    @staticmethod
    def get_isQ_dir():
        isqcBIN = os.popen("which isqc").read().strip()
        isqcDIR = os.path.dirname(isqcBIN)
        return isqcDIR

    @classmethod
    def get_simulatorBIN_path(cls):
        return os.path.join(cls.get_isQ_dir(), cls.set_simulator_fileName())

    @classmethod
    def makeSimulatorBIN(cls):
        # will automatically find dir of isqc, and make a SIMULATOR with appropreate chmod
        simBinFile = cls.get_simulatorBIN_path()
        if os.path.exists(simBinFile):
            return
        isQdir = cls.get_isQ_dir()
        messName = cls.get_dirName_of_Simulator(isQdir)
        replaceKeyWord = f"/nix/store/{messName}/bin/simulator"
        isqcFile = os.path.join(isQdir, "isqc")
        with open(isqcFile, "r") as f:
            isqcContent = f.read()
        slides = isqcContent.split(" ")
        slides[-2] = f"/nix/store/{messName}/bin/simulator"
        simBINcontent = " ".join(slides)
        with open(simBinFile, "w") as f:
            f.write(simBINcontent)
        os.chmod(simBinFile, 0o555)


class SIMULATOR(object):
    # base class

    def __init__(self, soFile, **kw):
        #         self.so = soFile
        self.kw = kw
        self.so = os.path.abspath(soFile)
        self.fileDir = os.path.dirname(self.so)
        self.SimulatorBIN = findSimulator.get_simulatorBIN_path()
        findSimulator.makeSimulatorBIN()

    def getsoDir(self):
        return os.path.dirname(self.so)

    def getSoFileName(self):
        return os.path.basename(self.so)

    def get_qcis_SO_fileName(self):
        return self.getSoFileName().replace(".so", "_QCIS_TMP_.so")

    def get_SIMULATOR(self):
        return self.SimulatorBIN

    def get_isQPath(self):
        return self.so.split(".so")[0] + ".isq"

    @staticmethod
    def bitfield(n: int, L: int):  # -> list[int]
        result = np.binary_repr(n, L)
        r = [int(digit) for digit in result]  # [2:] to chop off the "0b" part
        return r

    @staticmethod
    def calculate(func, res):
        # can calculate sum_i <i|A|i> = sum_i p_i * a_i ,
        # func(bitField) -> float,  bitField is in a reversed order (lower index first)
        r = 0
        for b in res:
            bitField = [int(i) for i in b]
            bitField.reverse()
            #             print(res,bitField,666)
            r += func(bitField) * res[b]
        return r


class isQSimulator(SIMULATOR):
    # run so file with isqc-simulator

    def runCircuit(self, F=[], I=[], shots=1000):
        # return {  "101":0.12, ...   }
        passF = "".join([f" -d {f}" for f in F])
        passI = "".join([f" -i {i}" for i in I])
        cmd = f"isqc simulate --shots {shots}  {passI} {passF} {self.so}"
        res = json.loads(os.popen(cmd).read())
        N = 0
        Res = {}
        for k in res:
            N += res[k]
        for k in res:
            #             Res[int(k,2)] = res[k]/N
            Res[k] = res[k] / N
        return Res


class FullSpaceSimulator(SIMULATOR):
    # compile each time, very slow

    def getisQfile(self):
        isQfile = self.so.split(".so")[0] + ".isq"
        return isQfile

    def getQCIScode(self, F=[], I=[]):
        isqFile = self.getisQfile()
        passF = "".join([f" -d {f}" for f in F])
        passI = "".join([f" -i {i}" for i in I])
        cmd = f"isqc compile {isqFile} --target qcis {passI} {passF}"
        os.system(cmd)
        return "tmp.qcis"

    def runCircuit(self, F=[], I=[], shots=1000):
        # return {  "101":0.12, ...   }
        isqf = self.getisQfile()
        with tempfile.TemporaryDirectory() as tmpdir:
            cwd = os.getcwd()
            os.chdir(tmpdir)
            passF = "".join([f" -d {f}" for f in F])
            passI = "".join([f" -i {i}" for i in I])
            cmd = f"isqc compile {isqf} --target qcis {passI} {passF}"
            os.system(cmd)
            os.chdir(cwd)
            qcis = isqf.split(".isq")[0] + ".qcis"
            #             backend = NumpyBackend()
            qc = IsqCircuit(
                file=qcis,
                backend=backend,
                sample=False,
            )
            r = {}
            M = qc.measure()
            for i in range(len(M)):
                r[bin(i)[2:]] = M[i]
            return r


class simulatorQICSbyYYS(SIMULATOR):
    def make_qcisSO(self, F=[], I=[]):
        # a fake so file is needed. This so file is a template for generating qcis
        fakeSO = os.path.join(self.getsoDir(), self.get_qcis_SO_fileName())
        # print("fakeSO=",fakeSO,666)
        if not os.path.exists(fakeSO):
            passF = "".join([f" -d {f}" for f in F])
            passI = "".join([f" -i {i}" for i in I])
            cmd = f"isqc compile {self.get_isQPath()} --target qcis {passI} {passF} -o {fakeSO}"
            os.system(cmd)
        return fakeSO

    def getQCIScode(self, F=[], I=[], **kw):
        fakeSO = self.make_qcisSO(F=F, I=I)
        simulatorBIN = self.get_SIMULATOR()
        passF = "".join([f" -d {f}" for f in F])
        passI = "".join([f" -i {i}" for i in I])
        cmd = f"{simulatorBIN} {fakeSO} --qcisgen -e __isq__entry {passI} {passF} "
        QCIS = os.popen(cmd).read()
        return QCIS

    def runCircuit(self, F=[], I=[], **kw):
        with open("tmp.qcis", "w") as f:
            f.write(self.getQCIScode(F=F, I=I))
        qc = IsqCircuit(
            file="tmp.qcis",
            backend=backend,
            sample=False,
        )
        r = {}
        M = qc.measure()
        # print(qc.state)
        os.system("rm tmp.qcis")
        nQ = int(
            np.log2(
                len(M) + 0.001,
            )
        )
        # print(66666,M)
        for i in range(len(M)):
            r[(f"{{:0{nQ}b}}").format(i)[::-1]] = M[i]
        #             r[ bin(i)[2:] ] = M[i]
        return r


class CircuitSimulatorByPharseQCIS:
    #  , ,,  'CY', 'CX', 'M', 'RXY'

    def pharseAll(self):
        lines = self.qcisStr.split("\n")
        for l in lines:
            l = l.strip()
            if not l:
                continue
            self.pharseOneLine(l.split(" "))

    # def pharseOneLine(self,segs):
    #     gateName = segs[0]
    #     if gateName in ['H', 'X', 'Y', 'Z', 'S', 'T','SD', 'TD','X2M', 'X2P', 'Y2M', 'Y2P']:
    #         # print(gateName,self.qdic[segs[1]],6666666)
    #         print(self.qdic,777)
    #         getattr( self.c , gateName )( self.qdic[segs[1]] )
    #     elif gateName in ['RZ', 'RX', 'RY']:
    #         getattr( self.c , gateName )( self.qdic[segs[1]] , float(segs[2])  )
    #     elif gateName in ['CZ','CNOT']:
    #         getattr( self.c , gateName )( self.qdic[segs[1]] , self.qdic[segs[2]]  )
    #     elif gateName == "M":
    #         pass
    #     else:
    #         logging.error(f"Undefined gatenName = {gateName}")

    def pharseOneLine(self, segs):
        gateName = segs[0]
        if gateName in [
            "H",
            "X",
            "Y",
            "Z",
            "S",
            "T",
            "SD",
            "TD",
            "X2M",
            "X2P",
            "Y2M",
            "Y2P",
        ]:
            getattr(self.c, gateName)(int(segs[1][1:]))
        elif gateName in ["RZ", "RX", "RY"]:
            getattr(self.c, gateName)(int(segs[1][1:]), float(segs[2]))
        elif gateName in ["CZ", "CNOT"]:
            getattr(self.c, gateName)(int(segs[1][1:]), int(segs[2][1:]))
        elif gateName == "M":
            pass
        else:
            logging.error(f"Undefined gatenName = {gateName}")

    @staticmethod
    def check(line_data):
        qdic = {}
        qnum = 0
        for idx, line in enumerate(line_data):
            line = line.strip()
            if not line:
                continue
            strArr = line.split(" ")

            if strArr[1] not in qdic:
                qdic[strArr[1]] = qnum
                qnum += 1

            if strArr[0] in ["CZ", "CY", "CX", "CNOT"]:
                if strArr[2] not in qdic:
                    qdic[strArr[2]] = qnum
                    qnum += 1
        return qnum, qdic

    def __init__(self, qcisStr, ctype="f", **kw):
        self.kw = kw
        self.qnum, self.qdic = self.check(qcisStr.split("\n"))
        if ctype == "f":
            self.c = Circuit.CircuitExF(self.qnum)
        else:
            self.c = Circuit.CircuitEx(self.qnum, IsSymbol=False)
        self.qcisStr = qcisStr

    def Measure(self):
        self.pharseAll()
        res = self.c.MeasureAll()
        if "printState" in self.kw:
            print("wavefunction:")
            for i in range(len(self.c.c)):
                if np.abs(self.c.c[i]) > 0.00001:
                    print(f"|{ (f'{{:0{self.c.getN()}b}}').format(i) }>", self.c.c[i])
        return res


class simulatorQICS_exact(SIMULATOR):
    def make_qcisSO(self, F=[], I=[]):
        # a fake so file is needed. This so file is a template for generating qcis
        fakeSO = os.path.join(self.getsoDir(), self.get_qcis_SO_fileName())
        if not os.path.exists(fakeSO):
            passF = "".join([f" -d {f}" for f in F])
            passI = "".join([f" -i {i}" for i in I])
            cmd = f"isqc compile {self.get_isQPath()} --target qcis {passI} {passF} -o {fakeSO}"
            os.system(cmd)
        return fakeSO

    def getQCIScode(self, F=[], I=[], **kw):
        simulatorBIN = self.get_SIMULATOR()
        passF = "".join([f" -d {f}" for f in F])
        passI = "".join([f" -i {i}" for i in I])
        cmd = (
            f"{simulatorBIN} {self.FAKE_SO} --qcisgen -e __isq__entry {passI} {passF} "
        )
        QCIS = os.popen(cmd).read()
        return QCIS

    def initCalc(self, nPara=1):
        self.FAKE_SO = self.make_qcisSO(F=[1.0] * nPara, I=[1] * nPara)

    def runCircuit(self, F=[], I=[], **kw):
        self.FAKE_SO = self.make_qcisSO(F=F, I=I)
        qcisStr = self.getQCIScode(F=F, I=I)
        c = CircuitSimulatorByPharseQCIS(qcisStr, **self.kw)
        res = c.Measure()
        r = {}
        for i in res:
            r[(f"{{:0{c.qnum}b}}").format(i)] = res[i]
        return r
