from abc import abstractmethod
import os 
import json 
import tempfile
import logging 

class BackendError(Exception):
    ''''''


class Backend():
    """Connect to backend, such as simulators and real devices.
    isq resource 

    Returns:
        _type_: _description_
    """    


    def __init__(self,circuit,workDir=None,**kw):          
        self.circuit = circuit 
        self.workDir = workDir
        self.isCreateIsQNeeded = None # flag to mark if recreate isq resource, children class may need this 
        self.kw = kw 
        if workDir is None:
            self.tmpDirObj = tempfile.TemporaryDirectory() 
            self.workDir = self.tmpDirObj.name
            self.isDelNeeded = True 
        else:
            self.workDir = workDir
            self.isDelNeeded = False 
            if not os.path.exists(self.workDir):
                os.makedirs(self.workDir)
        

    def __del__(self):
        if self.isDelNeeded:
            self.tmpDirObj.cleanup()


    def submit(self,paramList=None) -> list:
        """submit parameters (if there is any) into backend and return results. 

        Args:
            paramList (list, optional): element is dict ={'F':[],'I':[] } represents input args for circuit. Defaults to None.

        Returns:
            list: If paramList is None, return [result] (length of list is 1)
        """   
        # ----- create isq source code 
        self.make_isq_resource()     
        if self.circuit.isInputParameter(): 
            if paramList is None:
                WarnStr = f'''This is a parameterized circuit, a submission must contains input parameters. Details see docs: \n\n {self.submit.__doc__}'''
                logging.error(WarnStr) 
                raise BackendError(WarnStr) 
            else:
                '''run job with parameters'''
                logging.info(f"total num of jobs: {len(paramList)}")
                Return = [] # final return 
                truelySubmit = [] # for those on caches, submit 
                for i in range(len(paramList)):
                    para = paramList[i]
                    res = self.getCache( self._getParaKey(para) ) 
                    if res is None:
                        truelySubmit.append( para ) 
                        Return.append( len(truelySubmit) - 1 )
                    else:
                        Return.append( res )
                logging.info(f"[{len(truelySubmit)}] of them are not calculated") 
                calcResults = self.runJob_withParameters( paramList = truelySubmit )
                for i in range(len(Return)):
                    if type(Return[i]) is int:
                        Return[i] = calcResults[ Return[i] ]
                        self.setCache(key = self._getParaKey(paramList[i]), val= Return[i] )
                return Return
        else:
            if paramList is not None:
                logging.warning("This circuit needs not input parameters, the external input parameters are ignored") 
            else:
                '''run job without parameters'''
                result = self.runJob_noParameter()
                return [result]

    @abstractmethod
    def runJob_noParameter(self)->dict:
        '''
        '''

    @abstractmethod
    def runJob_withParameters(self,paramList)->list:
        '''
        '''

    def make_isq_resource(self):
        '''
        1. write isq file 
        2. compile to get .so file
        '''
        isqCode = self.circuit._makeCode() 
        isqFile = os.path.join(self.workDir,"resource.isq")
        resultsFile = os.path.join(self.workDir,'results.json') 
        if os.path.exists(isqFile):
            with open(isqFile,'r') as f:
                if f.read() == isqCode:
                    self.isCreateIsQNeeded = False 
                else:
                    self.isCreateIsQNeeded = True 
        else:
            self.isCreateIsQNeeded = True  
        if self.isCreateIsQNeeded:
            #>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
            #    put lib dir (isq lib for compiling)
            #---------------  need better design ----------------- 
            src_demoLib = os.path.join(  os.getcwd(), 'demoLib' )
            dst_demoLib = os.path.join( self.workDir, 'demoLib' )
            try:
                os.remove(dst_demoLib)
            except:
                pass 
            os.symlink(src=src_demoLib,dst=dst_demoLib)  
            #<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
            with open(isqFile,'w') as f:
                f.write(isqCode)
            tmp_cwd = os.getcwd()
            os.chdir( self.workDir )
            cmd = f"isqc compile {isqFile}"
            os.system(cmd)
            os.chdir(tmp_cwd) 
            with open(resultsFile,'w') as f:
                f.write("{}")


    def getCache(self,key):
        resultsFile = os.path.join(self.workDir,'results.json') 
        with open(resultsFile,'r') as f:
            d = json.loads(f.read())
        return d.get(key,None) 
    
    def setCache(self,key,val):
        resultsFile = os.path.join(self.workDir,'results.json') 
        with open(resultsFile,'r') as f:
            d = json.loads(f.read())
        d[key] = val 
        with open(resultsFile,'w') as f:
            f.write( json.dumps(d,indent=4) )

    @staticmethod
    def _getParaKey(paraDict):
        def singleFloatKey(f):
            return str(int(f * 100000000000000))
        intKey = "".join( [str(i) for i in paraDict['I'] ] )
        floKey = "".join( [ singleFloatKey(f)  for f in paraDict['F'] ] )
        return floKey + "_" + intKey




        

    

    
    # def make_isq_resource(self):
    #     workDir = self.workDir
    #     if workDir is None:
    #         with tempfile.TemporaryDirectory() as tmpdir:
    #             pass
    #             # self. 